# Momentum
`Momentum` is a widget that needs to be on the root of your app. It holds all the dependencies and configuration to be reused down the tree and also for configurations.

```dart
void main() {
  runApp(
    Momentum(
      // child: ...,
      // controllers: [...],
      // ...,
    ),
  );
}
```

!> Whenever you do changes in `Momentum` root widget like changing parameters, always `hot restart` the app instead of *hot reload*.

## child
- Category: `Constructor Parameter`
- Type: `Widget`
- Default: `null`
- Required: `YES`

The base or main entry widget for the app. This is usually`MyApp()`.
  ```dart
    Momentum(
      child: MyApp(),
      // controllers: [...],
      // ...,
    ),
  ```

<hr>

## controllers
- Category: `Constructor Parameter`
- Type: `List<MomentumController>`
- Default: `[]` *// empty list*
- Required: `YES`

This is where you instantiate all `MomentumController`. All controllers listed here will be used down the tree in both widgets and other controllers as well.
  ```dart
    Momentum(
      // child: ...,
      controllers: [
        ExampleControllerA(),
        ExampleControllerB(),
        ExampleControllerC(),
        // ...
      ],
      // ...,
    ),
  ```

<hr>

## services
- Category: `Constructor Parameter`
- Type: `List<MomentumService>`
- Default: `[]` *// empty list*
- Required: `NO`

This is where you instantiate all `MomentumServices`. All services listed here will be used down the tree inside widgets and inside controllers as well.
  ```dart
    Momentum(
      // child: ...,
      // controllers: [...],
      services: [
        ApiService(),
        DatabaseService(),
        SomeService(),
      ],
    ),
  ```

<hr>

## appLoader
- Category: `Constructor Parameter`
- Type: `Widget`
- Default: `null`
- Required: `NO`

Display custom loader widget that is used for awaiting all asynchronous initializers in your app before the `child` gets displayed. It's like a splash screen.
  ```dart
    Momentum(
      // child: ...,
      // controllers: [...],
      appLoader: MyAppLoaderWidget(),
    ),
  ```

<hr>

## disabledPersistentState
- Category: `Constructor Parameter`
- Type: `bool`
- Default: `false`
- Required: `NO`

Globally disable persistent state for all models. This is for models only not router. This will be also overridden by [skipPersist()](/momentum-controller?id=skippersist).
  ```dart
    Momentum(
      // child: ...,
      // controllers: [...],
      disabledPersistentState: true,
    ),
  ```

## lazy
- Category: `Constructor Parameter`
- Type: `bool`
- Default: `true`
- Required: `NO`

If set to `false`, all controllers will be bootstrapped when the app starts.
  ```dart
    Momentum(
      // child: ...,
      // controllers: [...],
      lazy: false,
    ),
  ```

<hr>

## maxTimeTravelSteps
- Category: `Constructor Parameter`
- Type: `int`
- Default: `1`
- Required: `NO`

Configure time travel feature. If the value is greater than `1`, the time travel feature will be enabled, otherwise, it is disabled. The maximum value is `250` steps.
  ```dart
    Momentum(
      // child: ...,
      // controllers: [...],
      maxTimeTravelSteps: 50, // 50 undo-redo steps only.
    ),
  ```

<hr>

## persistSave
- Category: `Constructor Parameter`
- Type: `Future<bool> Function(BuildContext, String, String)`
- Default: `null`
- Required: `NO`

A function parameter for saving or persisting state values using any storage provider. `key` and `value` are automatically generated by momentum. `BuildContext` is also provided so if your storage provider requires context, you can easily use it. This is a `Future<bool>` but you can use synchronous code here, just be sure that you return a bool value indicating whether key and value were saved successfully. `value` is a JSON representation of your model.
  ```dart
    Momentum(
      // child: ...,
      // controllers: [...],
      persistSave: (context, key, value) async {
        // save "key" and "value" somewhere.
        var sharedPref = SharedPreferences.getInstance();
        var success = await sharedPref.setString(key, value);
        return success;
      },
    ),
  ```

<hr>

## persistGet
- Category: `Constructor Parameter`
- Type: `Future<String> Function(BuildContext, String)`
- Default: `null`
- Required: `NO`

A function parameter for getting persisted state values using any storage provider. `key` is automatically generated by momentum. `BuildContext` is also provided so if your storage provider requires context, you can easily use it. This is a `Future<String>` but you can use synchronous code here. In the example code below,  `valueString` is a JSON representation of your model and momentum will automatically parse it.
  ```dart
    Momentum(
      // child: ...,
      // controllers: [...],
      persistGet: (context, key) async {
        // get data using the "key" provided.
        var sharedPref = SharedPreferences.getInstance();
        var valueString = sharedPref.getString(key);
        return valueString;
      },
    ),
  ```

<hr>

## onResetAll
- Category: `Constructor Parameter`
- Type: `void Function(BuildContext, void Function(BuildContext))`
- Default: `null`
- Required: `NO`

A function parameter for overriding `Momentum.resetAll(context)` method. You can use this to execute some logic before actually resetting all controllers.
  ```dart
    Momentum(
      // child: ...,
      // controllers: [...],
      onResetAll: (context, resetAll) async {
        var confirmed = await showDialog<bool>(context: context, builder: ...);
        if (confirmed) {
          // call the original resetAll method.
          resetAll(context);
        }
      },
    ),
  ```

<hr>

## enableLogging
- Category: `Constructor Parameter`
- Type: `bool`
- Default: `false`
- Required: `NO`

Whether to enable or disable logging for all controllers.
  ```dart
    Momentum(
      // child: ...,
      // controllers: [...],
      enableLogging: true, // enable logging for all controllers.
    ),
  ```

<hr>

## minimumBootstrapTime
- Category: `Constructor Parameter`
- Type: `int`
- Default: `0`
- Unit: `milliseconds`
- Required: `NO`

Forces the app to have a minimum set of delays before `child` will be displayed. It will display `appLoader` until the minimum time specified finishes. If asynchronous initializers completed longer than specified here, this parameter is ignored. This is in milliseconds.
  ```dart
    Momentum(
      // child: ...,
      // controllers: [...],
      minimumBootstrapTime: 3000, // minimum of 3 seconds loading.
    ),
  ```

<hr>

## .controller\<T\>(context)
- Category: `Static Method`
- Return Type: `T` *extends* `MomentumController`
- Parameters: `BuildContext`

Get a specific controller of type `T`. It requires context so you can only call this insides widgets.
  ```dart
    @override
    Widget build(BuildContext context) {
      var exampleController = Momentum.controller<ExampleController>(context);
      return Scaffold(...);
    }
  ```

<hr>

## .service\<T\>(context)
- Category: `Static Method`
- Return Type: `T` *extends* `MomentumService`
- Parameters: `BuildContext`

Get a specific service of type `T`. It requires context so you can only call this insides widgets.
  ```dart
    @override
    Widget build(BuildContext context) {
      var exampleService = Momentum.service<ExampleService>(context);
      return Scaffold(...);
    }
  ```

<hr>

## .restart(...)
- Category: `Static Method`
- Return Type: `void`
- Parameters: `BuildContext`, `Momentum`

Restart the current momentum app with a new momentum instance. Context is required.
  ```dart
    @override
    Widget build(BuildContext context) {
      return Scaffold(
        ...
          ...
            onPressed: () {
              Momentum.restart(context, momentum());
            },
      );
    }
  ```

<hr>

## .resetAll(context)
- Category: `Static Method`
- Return Type: `void`
- Parameters: `BuildContext`

Reset all controllers and their models. This will set all the models back to their initial values.
  ```dart
    @override
    Widget build(BuildContext context) {
      return Scaffold(
        ...
          ...
            onPressed: () {
              Momentum.resetAll(context);
            },
      );
    }
  ```

<hr>
